/*
 * Copyright (c) Alex Allan <alex X kamaz.org.uk> 
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <iostream>
#include <vector>
#include <string>
#include <algorithm>
#include <functional>

#include <taglib/fileref.h>
#include <taglib/tfile.h>

#include "cddb_query.h"
#include "tag.h"

using namespace TagLookup;

struct print_problemfile
    : public std::unary_function<const ProblemFile &, void>
{
    void operator() (const ProblemFile &pf)
    {
        TagLib::FileRef rf = pf.file;
        TagLib::File *f = rf.file();
        
        std::cout << "Problem File: " << f->name() << " Reason: " << FileDisc::ProblemTypeToString(pf.type) << std::endl;
    }
};

struct print_track 
    : public std::unary_function<const TagLookup::Track &, void>
{
    std::ostream &os;
    unsigned int s;
    
    print_track(std::ostream &out, unsigned int spaces) 
        : os(out), s(spaces)
    { }
    
    void operator() (const TagLookup::Track &track)
    {
        for(unsigned int i = 0; i < s; ++i)
            os << ' ';
        
        unsigned int n = track.getTrackNo();
        os << (n < 10 ? " " : "") << n << ". " << track.getName() << '\n'; 
    }
}; // print_track

struct print_disc 
    : public std::unary_function<const TagLookup::Disc &, void>
{
    std::ostream &os;
    unsigned int count;
    
    print_disc(std::ostream& out) 
        : os(out), count(0)
    {
    }
    
    void operator() (const TagLookup::Disc &disc)
    {
        ++count;
        
        os << '[' << (count < 10 ? " " : "") << count << "] " << disc.getArtist() << " - " << disc.getTitle() << '\n';
        
        TagLookup::Tracks ts = disc.getTracks(); 
        
        std::for_each(ts.begin(), ts.end(), print_track(os, 5));
    }
}; // print_disc


template<class T>
struct print
    : public std::unary_function<T, void>
{
    std::string p_;
    unsigned int c_;
    
    print(const std::string & prefix)
        : p_(prefix), c_(0)
    {
    }
    
    void operator() (T x)
    {
        std::cout << p_ << c_++ << ": " << x << std::endl;
    }
};

int main()
{
    std::vector<TagLib::FileRef> fl;
    
fl.push_back(TagLib::FileRef("/home/alex/simple/01.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/02.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/03.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/04.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/05.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/06.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/07.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/08.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/09.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/10.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/11.mp3"));
fl.push_back(TagLib::FileRef("/home/alex/simple/12.mp3"));

    FileDisc fd(fl);
    
    CDDBQuery q("misc", 0xd410710e);
    
    q.query();
    
    const Discs &d = q.getResults();

std::cout << "rah!" << std::endl;

    for_each(d.begin(), d.end(), print_disc(std::cout));
    
    ProblemFiles pfs = fd.validForUpdate();
    
    if(!pfs.empty())
        for_each(pfs.begin(), pfs.end(), print_problemfile());
    else
    {
        std::cout << "No Problems!" << std::endl;
        fd.tagByOrder(d.front());
    }
    return 0;
}

